/***************************************************************************
 *   Copyright (C) 2004-2006 by Jim Campbell                               *
 *   ifpgui@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "progressdlg.h"
#include <qapplication.h>
#include <qstring.h>
#include <qmessagebox.h>

// progess stuff
ProgressWindow *pgw;
int progress(void *context, struct ifp_transfer_status *status);

ProgressWindow::ProgressWindow(const char *caption)
    : QDialog(qApp->mainWidget(), "progress_dialog", FALSE,
             Qt::WStyle_StaysOnTop | Qt::WStyle_Tool)
{
	setCaption(caption);
	setWFlags(Qt::WStyle_StaysOnTop | Qt::WStyle_Tool);

	lastFileName = "";

	layout = new QVBoxLayout(this, 5, 4);

	fileprogress = new QProgressBar(25, this, "FileProgress");
	totalprogress = new QProgressBar(25, this, "TotalProgress");

	lblFilename = new QLabel(tr("FileName"), this);
	lblFilename->setAlignment(int(QLabel::AlignVCenter | QLabel::AlignLeft));
	lblFilename->setMinimumSize(400, 30);
	lblFilename->setMaximumSize(400, 30);

	btnCancel = new QPushButton(tr("Cancel"), this);
	btnCancel->setMaximumSize(200, 30);

	connect(btnCancel, SIGNAL(clicked()),
			  this, SLOT(btnCancel_clicked()));

	layout->addSpacing(10);
	layout->addWidget(lblFilename);
	layout->addWidget(fileprogress);
	layout->addSpacing(10);
	layout->addWidget(totalprogress);
	layout->addSpacing(15);
	layout->addWidget(btnCancel, 0, Qt::AlignHCenter);

	cancel_progress = false;

	setFileProgressTarget(0);
	setTotalProgressTarget(0);
	setFileProgress(0);
	setTotalProgress(0);

	this->setModal(FALSE);
}

/*
 *  Destroys the object and frees any allocated resources
 */
ProgressWindow::~ProgressWindow()
{
	 // no need to delete child widgets, Qt does it all for us
}


void ProgressWindow::btnCancel_clicked()
{
	//emit btnCancelClicked();
	//reject();
	cancel_progress = true;
}

/*!
	Creates a string with ... like "Trollte..." or "...olltech", depending on the alignment.
	I took this from the base qt source code.
*/
static QString qEllipsisText(const QString &org, const QFontMetrics &fm, int width, int align)
{
	int ellWidth = fm.width("...");
	QString text = QString::fromLatin1("");
	int i = 0;
	int len = org.length();
	int offset = (align & Qt::AlignRight) ? (len-1) - i : i;

	while (i < len && fm.width(text + org[offset]) + ellWidth < width)
		{
		if (align & Qt::AlignRight)
			text.prepend(org[offset]);
		else
			text += org[offset];
		offset = (align & Qt::AlignRight) ? (len-1) - ++i : ++i;
		}
	if (text.isEmpty())
		text = (align & Qt::AlignRight) ? org.right(1) : text = org.left(1);
	if (align & Qt::AlignRight)
		text.prepend("...");
	else
		text += "...";
	return text;
}

//=============================================================================
// take a file name and truncate it leaving the extension intact
//=============================================================================
QString ProgressWindow::truncateFileName(const QString &org, const QFontMetrics &fm, int width)
{
	int pos, extlen;
	QString tmpFileName = QString::fromLatin1("");
	QString fileExt = QString::fromLatin1("");
	int extWidth;

	// first extract the file extension from the filename
	pos = org.findRev('.');
	if (pos > 0)
		{
		// file name has an extension
		extlen = org.length() - pos;
		fileExt = org.mid(pos, extlen);
		tmpFileName = org.mid(0, pos);
		}
	else
		{
		// file name does not have an extension
		extlen = 0;
		tmpFileName = org;
		}
	extWidth = fm.width(fileExt);
	tmpFileName = qEllipsisText(tmpFileName, fm, width - extWidth, Qt::AlignLeft) + fileExt;
	return(tmpFileName);
}

//=============================================================================
// take a file path and remove whole parts replacing them with "..."
//=============================================================================
QString ProgressWindow::truncateFilePath(QString &fileStr)
{
	int pos;
	unsigned int cnt, middle;
	QStringList lst;
	QStringList::Iterator lst_it;
	QString truncFileName, tmpFileName, fileExt;

	pos = fileStr.findRev('\\');
	if (pos >= 0)
		lst = QStringList::split("\\", fileStr);
	else
		lst = QStringList::split("/", fileStr);

	// remove any "..." elements
	lst_it = lst.begin();
	while(lst_it != lst.end())
		{
		if (*lst_it == "...")
			lst_it = lst.remove(lst_it);
		else
			lst_it++;
		}

	middle = lst.size() / 2;

	// build truncated string
	if (pos >= 0)
		truncFileName = "\\";
	else
		truncFileName = "/";

	cnt = 1;
	lst_it = lst.begin();
	for ( ; lst_it != lst.end(); ++lst_it)
		{
		// replace middlemost element with a right side emphesis
		if (cnt != middle)
			{
			truncFileName += *lst_it;
			if (cnt != lst.size())
				{
				if (pos >= 0)
					truncFileName += "\\";
				else
					truncFileName += "/";
				}
			}
		else
			{
			truncFileName += "...";
			if (pos >= 0)
				truncFileName += "\\";
			else
				truncFileName += "/";
			}
		cnt++;
		}
	return(truncFileName);
}

void ProgressWindow::setFileName(const QString &fileName)
{
	int width, targetWidth;
	QString truncFileName, tmpFileName;

	if (fileName == lastFileName)
		return;

	tmpFileName = fileName;

	targetWidth = this->width() - (layout->margin() * 2);
	QFontMetrics fm = this->fontMetrics();
	width = fm.width(tmpFileName);
	if (targetWidth > 20)
		{
		do
			{
			truncFileName = truncateFilePath(tmpFileName);
			if (truncFileName == tmpFileName)
				{
				truncFileName = truncateFileName(tmpFileName, fm, targetWidth);
				}
			width = fm.width(truncFileName);
			tmpFileName = truncFileName;
			} while(width > targetWidth);
		}

	lblFilename->setText(truncFileName);
	lastFileName = truncFileName;
}

void ProgressWindow::setFileProgress(int progress)
{
	fileprogress->setProgress(progress);
}

void ProgressWindow::setFileProgressTarget(int totalSteps)
{
	fileprogress->setTotalSteps(totalSteps);
}

void ProgressWindow::setTotalProgress(int progress)
{
	if (totalprogress->isHidden() == FALSE)
		{
		totalprogress->setProgress(progress);
		}
}

void ProgressWindow::setTotalProgressTarget(int totalSteps)
{
	if (totalSteps == -1)
		{
		totalprogress->hide();
		}
	else
		{
		totalprogress->show();
		totalprogress->setTotalSteps(totalSteps);
		}
}

void ProgressWindow::hideCancelBtn(void)
{
	btnCancel->hide();
}

void ProgressWindow::showCancelBtn(void)
{
	btnCancel->show();
}

void ProgressWindow::removeStayOnTop(void)
{
	setWFlags(Qt::WStyle_Tool);
}

void ProgressWindow::updateProgress(ifp_progress_info *pg_info)
{
	setFileName(pg_info->file_name);

	if (pg_info->is_batch)
		{
		setFileProgressTarget(pg_info->file_total);
		setFileProgress(pg_info->file_bytes);

		if (pg_info->is_status_batch)
			{
			setTotalProgressTarget(pg_info->total_to_transfer);
			setTotalProgress(pg_info->total_transfered + pg_info->batch_bytes);// + pg_info->file_bytes);
			}
		else
			{
			setTotalProgressTarget(pg_info->total_to_transfer);
			setTotalProgress(pg_info->total_transfered + pg_info->file_bytes);
			}
		}
	else if (pg_info->is_status_batch)
		{
		setFileProgressTarget(pg_info->file_total);
		setFileProgress(pg_info->file_bytes);

		setTotalProgressTarget(pg_info->batch_total);
		setTotalProgress(pg_info->batch_bytes);// + pg_info->file_bytes);
		}
	else
		{
		setTotalProgressTarget(-1);

		setFileProgressTarget(pg_info->file_total);
		setFileProgress(pg_info->file_bytes);
		}
	qApp->processEvents();
}

int progress(void *context, struct ifp_transfer_status *status)
	{
	ifp_progress_info *pginfo;

	if (context == NULL)
		return(0); // this is really a bad error

	pginfo = (ifp_progress_info*)context;
	if (pginfo->pgw == NULL)
		return(0); // another bad error

	if (pginfo->pgw->cancel_progress == 1)
		return(1);

	pginfo->file_name		 = QString(status->file_name);
	pginfo->file_bytes		= status->file_bytes;
	pginfo->file_total		= status->file_total;
	pginfo->is_status_batch = status->is_batch;
	pginfo->batch_total	  = status->batch_total;
	pginfo->batch_bytes	  = status->batch_bytes;

	pginfo->pgw->updateProgress(pginfo);
	if (!pginfo->pgw->isShown())
		{
		pginfo->pgw->show();
		pginfo->pgw->raise();
		pginfo->pgw->setActiveWindow();
		}
	return(0);
	}
